# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.


from PySide6.QtCore import (
    Qt,
    QPoint,
    QPointF,
    QRectF,
)
from PySide6.QtGui import QFontMetrics
from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGraphicsRectItem,
    QGraphicsItem,
    QLabel,
    QLineEdit,
    QMenu,
    QMessageBox,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

from .constants import (
    TIMELINE_MASK_ON_COLOR,
    TIMELINE_MASK_OFF_COLOR,
    TIMELINE_TITLE_HEIGHT,
    TIMELINE_TITLE_BG_COLOR,
    TIMELINE_TITLE_FG_COLOR,
    TIMELINE_HEIGHT,
)
from .event import (
    EventCollection,
    ChangeEvent,
    ChooseEvent,
)
from .occurrence import Occurrence
from .textedit import TextEdit
from .utils import color_fg_from_bg


class Timeline(QGraphicsRectItem):

    def __init__(self, name, order, description, time_pane):
        super().__init__(time_pane)
        self.time_pane = time_pane
        self.event_collection = EventCollection()
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)
        self.setAcceptHoverEvents(True)
        self.title = TimelineTitle(name, self)
        self._name = None
        self.name = name
        self.mask = TimelineMask(self)
        self._select = False
        self._description = None
        self.description = description
        self.order = order

    @property
    def select(self):
        return self._select

    @select.setter
    def select(self, select):
        if select != self._select:
            self._select = select
            if select:
                self.time_pane.selected_timeline = self

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != self._name:
            self._name = name
            self.title.text = name

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description):
        if description != self._description:
            self._description = description
            if self._description == "":
                self._description = "(no description)"
            self.title.setToolTip(self._description)

    def occurrences(self):
        return sorted(
            [x for x in self.childItems() if isinstance(x, Occurrence)],
            key=lambda x: x.begin_time,
        )

    def update_rect(self):
        rect = QRectF(
            0,
            0,
            self.time_pane.scene.width(),
            TIMELINE_HEIGHT,
        )
        self.setRect(rect)
        self.mask.setRect(rect)

    def update_rect_width(self, new_width: float):
        """Update the width of the timeline"""
        rect = self.rect()
        rect.setWidth(new_width)
        self.setRect(rect)
        self.title.update_rect_width(new_width)
        self.mask.update_rect_width(new_width)

    def on_remove(self):
        answer = QMessageBox.StandardButton.Yes
        if len(self.occurrences()) > 0:
            answer = QMessageBox.question(
                self.time_pane.window,
                "Confirmation",
                "There are occurrences present. "
                "Do you want to remove this timeline?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
        if answer == QMessageBox.StandardButton.Yes:
            for ocr in self.occurrences():
                ocr.remove()
            self.time_pane.delete_timeline(self)

    def edit_properties(self):
        dialog = TimelinePropertiesDialog(self)
        dialog.exec()
        if dialog.result() == dialog.DialogCode.Accepted:
            is_changed = False
            name = dialog.get_name()
            if self.title.text != name:
                if not self.time_pane.check_new_timeline_name(name):
                    return
                self.name = name
                is_changed = True
            description = dialog.get_description()
            if self.description != description:
                self.description = description
                self.title.setToolTip(self.description)
                is_changed = True
            if is_changed:
                self.time_pane.data_needs_save = True

    def update_occurrences(self):
        for occurrence in self.occurrences():
            occurrence.update_style()

    def edit_events(self):
        while True:
            choose_dialog = ChooseEvent(
                self.event_collection, self.time_pane.view, "Finish"
            )
            choose_dialog.exec()
            if choose_dialog.result() == QMessageBox.DialogCode.Accepted:
                event = choose_dialog.get_chosen()
                name = event.name
                color = event.color
                description = event.description
                change_dialog = ChangeEvent(event, self.time_pane.window, True)
                change_dialog.exec()
                if (
                    change_dialog.result() == QMessageBox.DialogCode.Rejected
                    and change_dialog.request_remove()
                ):
                    self.remove_event(event)
                if (
                    name != event.name
                    or description != event.description
                    or color != event.color
                ):
                    self.update_occurrences()
                    self.time_pane.data_needs_save = True
            if choose_dialog.result() == QMessageBox.DialogCode.Rejected:
                break

        if self.event_collection.is_changed():
            self.time_pane.data_needs_save = True

    def remove_event(self, event):
        if event in set([x.event for x in self.occurrences()]):
            QMessageBox.warning(
                self.time_pane.window,
                "Warning",
                f'Event "{event.name}" cannot be removed because there are '
                "occurrences of it in the timeline",
            )
        else:
            self.event_collection.remove_event(event)

    def contextMenuEvent(self, event):
        menu = QMenu()
        menu.addAction(self.time_pane.window.menu.new_timeline_top_action)
        menu.addAction(self.time_pane.window.menu.new_timeline_above_action)
        menu.addAction(self.time_pane.window.menu.new_timeline_below_action)
        menu.addAction(self.time_pane.window.menu.new_timeline_bottom_action)
        menu.addAction("Delete timeline").triggered.connect(self.on_remove)
        menu.addAction("Edit timeline properties").triggered.connect(
            self.edit_properties
        )
        menu.addAction("Edit events").triggered.connect(self.edit_events)
        menu.addAction("Show events summary").triggered.connect(
            self.show_events_summary
        )
        try:
            pos = event.screenPos()
        except AttributeError:
            # Got here via keyPresEvent
            pos = QPoint(
                int(self.time_pane.window.width() / 2),
                int(self.time_pane.window.height() / 2),
            )
        menu.exec(pos)

    def new_timeline(self, where):
        self.time_pane.new_timeline(where)

    def can_create_occurrence(self, time):
        """Check whether an occurrence can be created at given time"""
        # Loop through the occurrences of the selected timeline
        valid = True
        for a in self.occurrences():
            if a.begin_time <= time <= a.end_time:
                valid = False
                break
        return valid

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Menu:
            self.contextMenuEvent(event)

    def show_events_summary(self):
        window = self.time_pane.window
        dialog = QDialog(window)
        dialog.setWindowTitle("Events summary")
        layout = QVBoxLayout()
        dialog.setLayout(layout)
        layout.addWidget(QLabel(f'Summary of events in timeline "{self.name}"'))
        table = QTableWidget(sum(1 for _ in self.event_collection), 4)
        table.setHorizontalHeaderLabels(
            ["events", "total time (s)", "count", "description"]
        )
        row = 0
        for name in self.event_collection:
            event = self.event_collection.get_event(name)
            bg_color = event.color
            fg_color = color_fg_from_bg(bg_color)
            item = QTableWidgetItem(name)
            item.setBackground(bg_color)
            item.setForeground(fg_color)
            item.setTextAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
            table.setItem(row, 0, item)
            count = 0
            total_time = 0
            for ocr in [x for x in self.occurrences() if x.event == event]:
                total_time += (ocr.end_time - ocr.begin_time) / 1000
                count += 1
            item = QTableWidgetItem(f"{total_time:.3f}")
            item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            table.setItem(row, 1, item)
            item = QTableWidgetItem(str(count))
            item.setTextAlignment(Qt.AlignCenter | Qt.AlignVCenter)
            table.setItem(row, 2, item)
            table.setItem(row, 3, QTableWidgetItem(event.description))
            row += 1
        table.horizontalHeader().setStretchLastSection(True)
        table.resizeColumnToContents(3)
        layout.addWidget(table)
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok,
            window,
        )
        buttons.accepted.connect(dialog.accept)
        layout.addWidget(buttons)
        dialog.exec()

    def select_occurrence(self, occurrence, direction):
        occurrences = self.occurrences()
        nb_occurrences = len(occurrences)
        if nb_occurrences > 0:
            idx = None
            for i, ocr in enumerate(occurrences):
                if ocr == occurrence:
                    idx = i
                    break
            occurrences[idx].setSelected(False)
            if direction == "next":
                idx += 1
                if idx == nb_occurrences:
                    idx = 0
            else:
                idx -= 1
                if idx < 0:
                    idx = nb_occurrences - 1
            occurrences[idx].setSelected(True)


class TimelineTitle(QGraphicsRectItem):

    def __init__(self, text: str, parent: Timeline = None):
        super().__init__(parent)
        self.text = text
        rect = self.parentItem().rect()
        rect.setHeight(TIMELINE_TITLE_HEIGHT)
        self.setRect(rect)
        self.parent = parent

    def paint(self, painter, option, widget=...):
        # Draw the rectangle
        self.draw_rect(painter)

        # Draw the text
        self.draw_text(painter)

    def draw_rect(self, painter):
        """Draw the timeline title rectangle"""
        # Set Pen and Brush for rectangle
        color = TIMELINE_TITLE_BG_COLOR
        painter.setPen(color)
        painter.setBrush(color)
        painter.drawRect(self.rect())

    def draw_text(self, painter):
        """Draw the timeline title text"""
        color = TIMELINE_TITLE_FG_COLOR
        painter.setPen(color)
        painter.setBrush(color)

        font = painter.font()
        fm = QFontMetrics(font)

        text_width = fm.boundingRect(self.text).width()
        text_height = fm.boundingRect(self.text).height()
        text_descent = fm.descent()

        # Get timeline polygon based on the viewport
        timeline_in_viewport_pos = self.parentItem().time_pane.view.mapToScene(
            self.rect().toRect()
        )

        bounding_rect = timeline_in_viewport_pos.boundingRect()

        # Get the viewport rect
        viewport_rect = self.parentItem().time_pane.view.viewport().rect()

        # Compute the x position for the text
        x_alignCenter = bounding_rect.x() + viewport_rect.width() / 2

        # No idea why the "-2", in the vertical position, is needed here
        text_position = QPointF(
            x_alignCenter - text_width / 2,
            TIMELINE_TITLE_HEIGHT / 2 + text_height / 2 - text_descent - 2,
        )

        painter.drawText(text_position, self.text)

    def update_rect_width(self, new_width):
        rect = self.rect()
        rect.setWidth(new_width)
        self.setRect(rect)


class TimelineMask(QGraphicsRectItem):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent

    def paint(self, painter, option, widget=...):
        if self.parent.select:
            self.setZValue(-1)
            color = TIMELINE_MASK_ON_COLOR
        else:
            self.setZValue(1)
            color = TIMELINE_MASK_OFF_COLOR
            painter.setPen(color)
            painter.setBrush(color)
            painter.drawRect(self.rect())

    def update_rect_width(self, new_width):
        rect = self.rect()
        rect.setWidth(new_width)
        self.setRect(rect)

    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            if not self.parent.time_pane.occurrence_under_creation:
                self.parent.time_pane.select_timeline(self.parent)
        event.ignore()


class TimelinePropertiesDialog(QDialog):

    def __init__(self, timeline):
        window = timeline.time_pane.window
        super().__init__(window)
        self.setWindowTitle("Timeline properties")
        layout = QFormLayout()
        self.name_edit = QLineEdit()
        self.name_edit.setText(timeline.name)
        self.name_edit.returnPressed.connect(self.accept)
        layout.addRow("Name: ", self.name_edit)
        layout.addRow("Description:", None)
        self.description_edit = TextEdit(self, timeline.description)
        layout.addRow(self.description_edit)
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel,
            window,
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)
        self.setLayout(layout)

    def get_name(self):
        return self.name_edit.text()

    def get_description(self):
        return self.description_edit.toPlainText()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
from math import (
    inf,
    isinf,
)

from PySide6.QtCore import Qt
from PySide6.QtGui import (
    QPainter,
    QColor,
)
from PySide6.QtMultimedia import QMediaPlayer
from PySide6.QtWidgets import (
    QAbstractSlider,
    QGraphicsRectItem,
    QGraphicsScene,
    QGraphicsView,
    QMessageBox,
    QScrollBar,
    QStyle,
)
from .constants import (
    CSV_HEADERS,
    EVENT_DEFAULT_COLOR,
    MINIMUM_ZOOM_FACTOR,
    OCCURRENCE_PEN_WIDTH_OFF_CURSOR,
    OCCURRENCE_PEN_WIDTH_ON_CURSOR,
    TIMELINE_HEIGHT,
    TIME_SCALE_HEIGHT,
    ZOOM_STEP,
)
from .event import (
    ChooseEvent,
    Event,
    EventCollection,
)
from .occurrence import (
    Occurrence,
    OccurrenceHandle,
)
from .timeline import (
    Timeline,
    TimelinePropertiesDialog,
)
from .timescale import TimeScale


class TimePaneView(QGraphicsView):

    def __init__(self, window):
        super().__init__(window)
        self.zoom_step = ZOOM_STEP
        self.zoom_shift = None
        self.minimum_zoom_factor = MINIMUM_ZOOM_FACTOR
        self.zoom_factor = self.minimum_zoom_factor
        self.window = window
        self.create_ui()

    def create_ui(self):
        vertical_scrollbar = QScrollBar(Qt.Orientation.Vertical, self)
        vertical_scrollbar.valueChanged.connect(
            self.on_vertical_scroll_value_changed
        )
        self.setVerticalScrollBar(vertical_scrollbar)

        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setFrameShape(QGraphicsView.Shape.NoFrame)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        # This is necessary for getting the cursor being updated
        self.setViewportUpdateMode(
            QGraphicsView.ViewportUpdateMode.FullViewportUpdate
        )
        self.scene = TimePaneScene(self)
        self.scene.setSceneRect(0, 0, self.width(), self.height())
        self.setScene(self.scene)

    def wheelEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            if not self.window.has_media():
                return
            mouse_pos = self.mapToScene(event.position().toPoint()).x()
            if event.angleDelta().y() > 0:
                self.zoom_shift = mouse_pos * (1 - self.zoom_step)
                self.zoom_in()
            else:
                self.zoom_shift = mouse_pos * (1 - 1 / self.zoom_step)
                self.zoom_out()
            self.zoom_shift = None
        elif event.modifiers() == Qt.KeyboardModifier.ShiftModifier:
            if event.angleDelta().y() > 0:
                action = QAbstractSlider.SliderSingleStepAdd
            else:
                action = QAbstractSlider.SliderSingleStepSub
            self.horizontalScrollBar().triggerAction(action)
        else:
            super().wheelEvent(event)

    def on_vertical_scroll_value_changed(self, value):
        """Allow the time scale to be always visible when scrolling"""
        if self.scene.time_pane.time_scale:
            self.scene.time_pane.time_scale.setPos(0, value)

    def zoom_in(self):
        self.zoom_factor *= self.zoom_step
        self.update_scale()

    def zoom_out(self):
        new_zoom_factor = self.zoom_factor / self.zoom_step
        if new_zoom_factor >= self.minimum_zoom_factor:
            self.zoom_factor = new_zoom_factor
            self.update_scale()

    def update_scale(self):
        # Update the size of the scene with zoom_factor
        self.scene.setSceneRect(
            0,
            0,
            self.width() * self.zoom_factor,
            self.scene.height(),
        )

        if self.zoom_shift:
            previous_anchor = self.transformationAnchor()
            self.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.translate(self.zoom_shift, 0)
            self.setTransformationAnchor(previous_anchor)

    def resizeEvent(self, a0):
        # FIXME: Instead of testing the presence of the time scale, test whether there is a video loaded
        if self.scene.time_pane.time_scale:
            origin = self.mapToScene(0, 0).x()
            width_before = self.scene.width() / self.zoom_factor
            width_after = self.width()
            shift = origin * (1 - width_after / width_before)
            self.update_scale()
            previous_anchor = self.transformationAnchor()
            self.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.translate(shift, 0)
            self.setTransformationAnchor(previous_anchor)
        else:
            self.scene.setSceneRect(
                0,
                0,
                self.width(),
                self.scene.height(),
            )
        self.update()
        super().resizeEvent(a0)


class TimePaneScene(QGraphicsScene):

    def __init__(self, view):
        super().__init__()
        self.view = view
        self.sceneRectChanged.connect(self.on_scene_changed)
        self.create_time_pane()

    def create_time_pane(self):
        self.time_pane = TimePane(self)
        self.time_pane.setY(TIME_SCALE_HEIGHT)
        self.addItem(self.time_pane)

    def on_scene_changed(self, rect):
        self.time_pane.on_change(rect)


class TimePane(QGraphicsRectItem):

    def __init__(self, scene):
        """Initialize the time pane graphics item"""
        super().__init__()
        self._duration = 0
        self._time = 0

        self.selected_timeline = None
        self.occurrence_under_creation: Occurrence = None
        self.scene = scene
        self.view = scene.view
        self.window = self.view.window
        self.time_scale = None
        self.data_needs_save = False
        self.scrollbar_width = self.window.style().pixelMetric(
            QStyle.PM_ScrollBarExtent
        )

    def on_change(self, rect):
        # Update occurrences
        for timeline in self.timelines():
            timeline.update_rect_width(rect.width())
            for occurrence in timeline.occurrences():
                occurrence.update_rect()

        if self.occurrence_under_creation:
            self.occurrence_under_creation.update_rect()

        # Update time scale display
        if self.time_scale:
            # Update cursor
            if self.duration:
                self.set_cursor_position(self.time, rect.width())
            self.time_scale.update_rect()

    def set_cursor_position(self, time, width):
        self.time_scale.cursor.setX(
            (time + self.window.frame_duration() / 2) * width / self.duration
        )

    def select_cycle_timeline(self, delta):
        timelines = self.timelines()
        i, n = self.find_selected_timeline()
        selected_timeline = timelines[i]
        selected_timeline.select = False
        if delta > 0:
            if i == n - 1:
                i = -1
        else:
            if i == 0:
                i = n
        i += delta
        self.select_timeline(timelines[i])

    def find_selected_timeline(self):
        timelines = self.timelines()
        n = len(timelines)
        for i in range(n):
            if timelines[i].select:
                break
        return i, n

    def select_timeline(self, line):
        for tl in self.timelines():
            tl.select = False
        line.select = True
        self.occurrence_borders()
        line.update()

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, time):
        time = self.window.disctretize_time(time)
        if time != self._time:
            self._time = time
            if self._time <= 0:
                self._time = 0
            max_duration = self.duration - self.window.frame_duration()
            if self._time > max_duration:
                self._time = max_duration

            # First, update the occurrence under creation, if it exists. If the
            # cursor time goes beyond the allowed bounds, bring it back and do
            # not update the other widgets.
            ouc = self.occurrence_under_creation
            if ouc:
                if ouc.lower_bound and self._time < ouc.lower_bound:
                    self._time = ouc.lower_bound
                elif ouc.upper_bound and self._time > ouc.upper_bound:
                    self._time = ouc.upper_bound
                    if (
                        self.window.video.playback_state()
                        == QMediaPlayer.PlaybackState.PlayingState
                    ):
                        self.window.media_player_pause()
                begin_time = ouc.begin_time
                frame_duration = self.window.frame_duration()
                if self._time >= begin_time:
                    ouc.update_end_time(self._time + frame_duration)
                else:
                    ouc.update_end_time(self._time)

            self.window.video.position = self._time

            # Update cursor position
            if self.time_scale.cursor:
                self.set_cursor_position(self._time, self.scene.width())

            if isinstance(self.scene.focusItem(), OccurrenceHandle):
                occurrence_handle: OccurrenceHandle = self.scene.focusItem()
                occurrence_handle.change_time(self._time)

            self.occurrence_borders()
            self.view.update()

    def occurrence_borders(self):
        # Change appearance of occurrence under the cursor
        # (Brute force approach; this ought to be improved)
        if not self.occurrence_under_creation:
            for tml in self.timelines():
                for ocr in tml.occurrences():
                    ocr.pen_width = OCCURRENCE_PEN_WIDTH_OFF_CURSOR
                    if tml.select:
                        hfd = self.window.frame_duration() / 2
                        if (
                            self.time + hfd > ocr.begin_time
                            and self.time + hfd < ocr.end_time
                        ):
                            ocr.pen_width = OCCURRENCE_PEN_WIDTH_ON_CURSOR

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, duration):
        if duration != self._duration:
            self._duration = duration
            # Recreate time pane scale and the cursor
            self.time_scale = TimeScale(self)
            self.scene.addItem(self.time_scale)
            self.view.update()

    def clear(self):
        # Clear timelineScene
        self.scene.clear()

    def handle_occurrence(self):
        """Handle the occurrence"""
        menu = self.window.menu
        if self.occurrence_under_creation is None:
            if self.selected_timeline.can_create_occurrence(self.time):
                frame_duration = self.window.frame_duration()
                self.occurrence_under_creation = Occurrence(
                    self.selected_timeline,
                    self.time,
                    self.time + frame_duration,
                )
            menu.start_occurence()
        else:
            # Finish the current occurrence
            events_dialog = ChooseEvent(
                self.selected_timeline.event_collection, self.view
            )
            events_dialog.exec()
            if events_dialog.result() == QMessageBox.DialogCode.Accepted:
                event = events_dialog.get_chosen()
                self.occurrence_under_creation.set_event(event)
                self.occurrence_under_creation.finish_creation()
                self.occurrence_under_creation = None
                self.occurrence_borders()
            menu.finish_occurence()
            self.data_needs_save = True
            self.view.update()

    def new_timeline(self, where):
        if not isinf(where):
            order = self.selected_timeline.order
            if where < 0:
                where = order - 0.5
            else:
                where = order + 0.5
        timeline = Timeline("", where, "", self)
        dialog = TimelinePropertiesDialog(timeline)
        dialog.exec()
        if dialog.result() == QMessageBox.DialogCode.Accepted:
            name = dialog.get_name()
            if not self.check_new_timeline_name(name):
                self.delete_timeline(timeline)
                return
            timeline.name = name
            timeline.description = dialog.get_description()
            self.add_timeline(timeline)
            self.place_timelines()
            self.data_needs_save = True

    def check_new_timeline_name(self, name):
        if name == "":
            QMessageBox.warning(
                self.window, "Warning", "Timeline name cannot be empty"
            )
            return False
        if name in self.get_timeline_names():
            QMessageBox.warning(
                self.window,
                "Warning",
                f'A timeline with name "{name}" exists already',
            )
            return False
        return True

    def abort_occurrence_creation(self):
        if self.occurrence_under_creation is not None:
            confirm_box = ConfirmMessageBox(
                "Abort creation of occurrence?", self.window
            )
            if confirm_box.result() == ConfirmMessageBox.DialogCode.Accepted:
                self.occurrence_under_creation.remove()
                self.occurrence_under_creation = None
                self.view.update()
                self.window.menu.start_occurence()

    def add_timeline(self, timeline):
        # Set the timeline rectangle
        timeline.update_rect()

        # Select the new timeline
        for i in self.timelines():
            i.select = False
        timeline.select = True

        self.view.update_scale()

    def place_timelines(self):
        order = 0
        for i in self.timelines():
            i.order = order
            order += 1
        timelines = self.timelines()
        for timeline in timelines:
            timeline.setPos(0, timeline.order * TIMELINE_HEIGHT)

        # Adjust the height of of the scene
        rect = self.scene.sceneRect()
        height = len(self.timelines()) * TIMELINE_HEIGHT + TIME_SCALE_HEIGHT
        rect.setHeight(height)
        self.scene.setSceneRect(rect)

        # Set maximum height of the widget
        self.view.setMaximumHeight(int(height) + self.scrollbar_width + 2)

    def move_timeline(self, delta):
        if isinf(delta):
            if delta < 0:
                self.selected_timeline.order = -1
            else:
                self.selected_timeline.order = len(self.timelines())
        else:
            self.selected_timeline.order += delta * 1.5
        self.place_timelines()

    def add_data(self, data):
        for _, row in data.iterrows():
            # Search for timeline
            timeline = self.get_timeline_by_name(row["timeline"])

            # If timeline from csv doesn't exist in TimePane,
            # escape row
            if not timeline:
                continue

            # Search for event
            event = timeline.event_collection.get_event(row["event"])

            # If event from csv doesn't exist in timeline,
            # then add it
            if not event:
                continue

            occurrence = Occurrence(
                timeline,
                int(row["begin"]),
                int(row["end"]),
            )

            occurrence.set_event(event)
            occurrence.finish_creation()

    def get_timeline_by_name(self, name):
        """Get the timeline by name"""
        return next((x for x in self.timelines() if x.name == name), None)

    def has_occurrences(self) -> bool:
        return any(len(line.occurrences()) for line in self.timelines())

    def delete_occurrence(self):
        for i in self.scene.selectedItems():
            if isinstance(i, Occurrence):
                i.on_remove()
                break

    def timelines_from_config(self, config):
        if "timelines" in config:

            # Set all absent order fields with Inf
            for k, v in config["timelines"].items():
                if not v:
                    v = dict()
                    config["timelines"][k] = v
                if "order" not in v:
                    v["order"] = -inf
                if "description" not in v:
                    v["description"] = ""

            # Sort according to order first and alphabetically from
            # timeline name, otherwise from the "order" property In the
            # loop below, the "order" attribute of the Timeline items will
            # receiving increasing nteger values, starting at zero.
            order = 0
            for item in sorted(
                config["timelines"].items(),
                key=lambda x: (x[1]["order"], x[0]),
            ):
                # Get name and properties of the timeline
                name = item[0]
                properties = item[1]
                description = properties["description"]

                # Create timeline
                line = Timeline(name, order, description, self)
                order += 1

                # Add the timeline to the TimePane
                self.add_timeline(line)

                # Loop over events of the timeline
                event_collection = EventCollection()
                if "events" in properties:
                    for k, v in properties["events"].items():
                        event = Event(k)
                        try:
                            event.color = QColor(v["color"])
                        except KeyError:
                            event.color = EVENT_DEFAULT_COLOR
                        try:
                            event.description = v["description"]
                        except KeyError:
                            event.description = ""
                        event_collection.add_event(event)
                    line.event_collection = event_collection

            self.place_timelines()

    def timelines_to_dataframe(self):
        df = pd.DataFrame(columns=CSV_HEADERS)
        for timeline in sorted(self.timelines(), key=lambda x: x.name):
            for occurrence in timeline.occurrences():
                comment = occurrence.comment.replace('"', '\\"')
                row = [
                    timeline.name,
                    occurrence.event.name,
                    f"{occurrence.begin_time:.3f}",
                    f"{occurrence.end_time:.3f}",
                    comment,
                ]
                df.loc[len(df.index)] = row
        return df

    def timelines(self):
        # Always return a list sorted by order
        return sorted(
            [x for x in self.childItems() if isinstance(x, Timeline)],
            key=lambda x: x.order,
        )

    def delete_timeline(self, timeline):
        # If the timeline to delted is the selected one, select the next one
        if self.selected_timeline == timeline:
            timelines = self.timelines()
            new_selected = None
            for i in timelines:
                if i.order > timeline.order:
                    new_selected = i
                    break
            if not new_selected:
                new_selected = timelines[0]
            new_selected.select = True
        self.scene.removeItem(timeline)
        self.place_timelines()
        del timeline
        self.data_needs_save = True

    def get_timeline_names(self):
        return [x.name for x in self.timelines()]

    def select_occurrence(self, direction):
        selected_items = self.scene.selectedItems()
        if len(selected_items) > 0:
            occurrence = selected_items[0]
            timeline = occurrence.timeline
            timeline.select_occurrence(occurrence, direction)


class ConfirmMessageBox(QMessageBox):
    def __init__(self, message, parent=None):
        super().__init__(
            QMessageBox.Icon.Warning,
            "Warning",
            message,
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            parent,
        )

        self.button(QMessageBox.StandardButton.Yes).clicked.connect(self.accept)
        self.button(QMessageBox.StandardButton.No).clicked.connect(self.reject)
        self.exec()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import platform
from functools import partial
from math import inf

from PySide6.QtCore import (
    Qt,
    QUrl,
)
from PySide6.QtGui import (
    QAction,
    QDesktopServices,
    QKeySequence,
)
from PySide6.QtWidgets import (
    QMessageBox,
    QStyle,
)

from .about import About
from .constants import (
    PYPI_URL,
    REPOSITORY_URL,
)


class Menu:

    def __init__(self, window):

        self.window = window
        self.menu_bar = window.menuBar()
        self.style = self.window.style()

        self.create_file_menu()
        self.create_play_menu()
        self.create_edit_menu()
        self.create_view_menu()
        self.create_data_menu()
        self.create_help_menu()

    def create_file_menu(self):
        menu = self.menu_bar.addMenu("&File")

        # Add actions to file menu
        open_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "&Open video",
            self.window,
            shortcut=QKeySequence.StandardKey.Open,
            triggered=self.window.files.open_video,
        )

        open_project_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "Open &project",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_O
            ),
            triggered=self.window.files.open_project,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_DialogSaveButton
        )
        self.save_project_action = QAction(
            icon,
            "Save project",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_S),
            triggered=self.window.files.save_project,
            enabled=False,
        )

        export_data_action = QAction(
            icon,
            "Export Data",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_S
            ),
            triggered=self.window.files.export_data_file,
        )

        export_config_action = QAction(
            icon,
            "Export Configuration",
            self.window,
            triggered=self.window.files.save_config_file,
        )

        close_action = QAction(
            "Quit",
            self.window,
            shortcut=(
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Q)
                if platform.system() == "Windows"
                else QKeySequence.StandardKey.Quit
            ),
            triggered=self.window.close,
        )

        menu.addAction(open_action)
        menu.addAction(open_project_action)
        menu.addAction(self.save_project_action)
        menu.addAction(export_data_action)
        menu.addAction(export_config_action)
        menu.addAction(close_action)

    def create_play_menu(self):
        # Add actions to play menu
        menu = self.menu_bar.addMenu("&Play")
        icon = self.style.standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        self.play_action = QAction(
            icon,
            "Play/Pause",
            self.window,
            shortcut=Qt.Key.Key_Space,
            triggered=self.window.video.play_pause,
            enabled=False,
        )

        icon = self.style.standardIcon(QStyle.StandardPixmap.SP_MediaStop)
        self.stop_action = QAction(
            icon,
            "Stop",
            self.window,
            shortcut=Qt.Key.Key_S,
            triggered=self.window.video.stop,
            enabled=False,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_MediaSkipBackward
        )
        self.previous_frame_action = QAction(
            icon,
            "Go to the previous frame",
            self.window,
            shortcut=Qt.Key.Key_Left,
            triggered=partial(self.window.move_to_frame, -1),
            enabled=False,
        )
        self.fifth_previous_frame_action = QAction(
            icon,
            "Go to the fifth previous frame",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Left),
            triggered=partial(self.window.move_to_frame, -5),
            enabled=False,
        )
        self.tenth_previous_frame_action = QAction(
            icon,
            "Go to the tenth previous frame",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Left
            ),
            triggered=partial(self.window.move_to_frame, -10),
            enabled=False,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_MediaSkipForward
        )
        self.next_frame_action = QAction(
            icon,
            "Go to the next frame",
            self.window,
            shortcut=Qt.Key.Key_Right,
            triggered=partial(self.window.move_to_frame, 1),
            enabled=False,
        )
        self.fifth_next_frame_action = QAction(
            icon,
            "Go to the fifth next frame",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Right),
            triggered=partial(self.window.move_to_frame, 5),
            enabled=False,
        )
        self.tenth_next_frame_action = QAction(
            icon,
            "Go to the tenth next frame",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Right
            ),
            triggered=partial(self.window.move_to_frame, 10),
            enabled=False,
        )

        for action in [
            self.play_action,
            self.stop_action,
            self.tenth_previous_frame_action,
            self.fifth_previous_frame_action,
            self.previous_frame_action,
            self.next_frame_action,
            self.fifth_next_frame_action,
            self.tenth_next_frame_action,
        ]:
            menu.addAction(action)

    def create_edit_menu(self):
        menu = self.menu_bar.addMenu("&Edit")

        # Edit Timeline submenu
        edit_timeline_menu = menu.addMenu("&Timeline")

        self.new_timeline_top_action = QAction(
            "Add new timeline at the top",
            self.window,
            shortcuts=[
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Return),
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Enter),
            ],
            triggered=partial(self.new_timeline, -inf),
            enabled=True,
        )
        edit_timeline_menu.addAction(self.new_timeline_top_action)

        self.new_timeline_above_action = QAction(
            "Add new timeline above selected",
            self.window,
            triggered=partial(self.new_timeline, -1),
            enabled=True,
        )
        edit_timeline_menu.addAction(self.new_timeline_above_action)

        self.new_timeline_below_action = QAction(
            "Add new timeline below selected",
            self.window,
            triggered=partial(self.new_timeline, 1),
            enabled=True,
        )
        edit_timeline_menu.addAction(self.new_timeline_below_action)

        self.new_timeline_bottom_action = QAction(
            "Add new timeline at the bottom",
            self.window,
            triggered=partial(self.new_timeline, inf),
            enabled=True,
        )
        edit_timeline_menu.addAction(self.new_timeline_bottom_action)

        self.edit_timeline_properties_action = QAction(
            "Edit Properties of Selected Timeline",
            self.window,
            triggered=self.edit_timeline_properties,
            enabled=True,
        )
        edit_timeline_menu.addAction(self.edit_timeline_properties_action)

        self.edit_timeline_events_action = QAction(
            "Edit Events of Selected Timeline",
            self.window,
            triggered=self.edit_timeline_events,
            enabled=True,
        )
        edit_timeline_menu.addAction(self.edit_timeline_events_action)

        select_next_timeline_action = QAction(
            "Select Next Timeline",
            self.window,
            shortcut=Qt.Key.Key_Down,
            triggered=partial(self.select_cycle_timeline, 1),
            enabled=True,
        )
        edit_timeline_menu.addAction(select_next_timeline_action)

        select_previous_timeline_action = QAction(
            "Select Previous Timeline",
            self.window,
            shortcut=Qt.Key.Key_Up,
            triggered=partial(self.select_cycle_timeline, -1),
            enabled=True,
        )
        edit_timeline_menu.addAction(select_previous_timeline_action)

        move_timeline_top_action = QAction(
            "Move Timeline To Top",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Key.Key_Up
            ),
            triggered=partial(self.move_timeline, -inf),
            enabled=True,
        )
        edit_timeline_menu.addAction(move_timeline_top_action)

        move_timeline_up_action = QAction(
            "Move Timeline Up",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Up),
            triggered=partial(self.move_timeline, -1),
            enabled=True,
        )
        edit_timeline_menu.addAction(move_timeline_up_action)

        move_timeline_down_action = QAction(
            "Move Timeline Down",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Down),
            triggered=partial(self.move_timeline, 1),
            enabled=True,
        )
        edit_timeline_menu.addAction(move_timeline_down_action)

        move_timeline_bottom_action = QAction(
            "Move Timeline To The Bottom",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Key.Key_Down
            ),
            triggered=partial(self.move_timeline, inf),
            enabled=True,
        )
        edit_timeline_menu.addAction(move_timeline_bottom_action)

        # Edit Occurrence submenu
        edit_occurrence_menu = menu.addMenu("&Occurrence")

        self.add_occurrence_action = QAction(
            "Start Occurrence",
            self.window,
            shortcuts=[Qt.Key.Key_Return, Qt.Key.Key_Enter],
            triggered=self.add_occurrence,
            enabled=False,
        )
        edit_occurrence_menu.addAction(self.add_occurrence_action)

        self.abort_occurrence_creation_action = QAction(
            "Abort Creation of Occurrence",
            self.window,
            shortcut=Qt.Key.Key_Escape,
            triggered=self.abort_occurrence_creation,
            enabled=False,
        )
        edit_occurrence_menu.addAction(self.abort_occurrence_creation_action)

        self.delete_occurrence_action = QAction(
            "Delete Occurrence",
            self.window,
            shortcuts=[Qt.Key.Key_Backspace, Qt.Key.Key_Delete],
            triggered=self.delete_occurrence,
            enabled=True,
        )
        edit_occurrence_menu.addAction(self.delete_occurrence_action)

        self.select_previous_occurence_action = QAction(
            "Select Previous Occurrence ",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.ALT | Qt.Key.Key_Left),
            triggered=partial(self.select_occurrence, "previous"),
            enabled=True,
        )
        edit_occurrence_menu.addAction(self.select_previous_occurence_action)

        self.select_next_occurence_action = QAction(
            "Select Next Occurrence ",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.ALT | Qt.Key.Key_Right),
            triggered=partial(self.select_occurrence, "next"),
            enabled=True,
        )
        edit_occurrence_menu.addAction(self.select_next_occurence_action)

    def new_timeline(self, where, checked=False):
        self.window.time_pane.new_timeline(where)

    def add_occurrence(self):
        self.window.time_pane.handle_occurrence()

    def select_cycle_timeline(self, delta, checked=False):
        self.window.time_pane.select_cycle_timeline(delta)

    def move_timeline(self, delta, checked=False):
        self.window.time_pane.move_timeline(delta)

    def abort_occurrence_creation(self):
        self.window.time_pane.abort_occurrence_creation()

    def delete_occurrence(self):
        self.window.time_pane.delete_occurrence()

    def create_data_menu(self):
        menu = self.menu_bar.addMenu("&Data")
        csv_delimiter_action = QAction(
            "Change CSV Delimiter",
            self.window,
            triggered=self.window.files.change_csv_delimiter,
        )
        menu.addAction(csv_delimiter_action)
        set_coder_action = QAction(
            "Set Coder",
            self.window,
            triggered=self.set_coder,
        )
        menu.addAction(set_coder_action)
        self.video_inclusion_exclusion_action = QAction(
            "Exclude video from project file at next save",
            self.window,
            triggered=self.window.files.toggle_video_inclusion,
        )
        menu.addAction(self.video_inclusion_exclusion_action)

    def set_coder(self):
        self.window.files.coders.set_current()

    def select_occurrence(self, direction, checked=False):
        self.window.time_pane.select_occurrence(direction)

    def create_view_menu(self):
        # Add actions to view menu
        menu = self.menu_bar.addMenu("&View")
        self.fullscreen_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_TitleBarMaxButton),
            "Toggle Fullscreen",
            self.window,
            shortcut=Qt.Key.Key_F11,
            triggered=self.window.on_fullscreen,
        )

        menu.addAction(self.fullscreen_action)

    def create_help_menu(self):
        # Help menu
        menu = self.menu_bar.addMenu("&Help")

        self.about_action = QAction(
            "About ViCodePy",
            self.window,
            triggered=self.about,
        )
        menu.addAction(self.about_action)

        self.visit_pypi_action = QAction(
            "Visit PyPI project",
            self.window,
            triggered=partial(self.open_url, PYPI_URL),
        )
        menu.addAction(self.visit_pypi_action)

        self.visit_repository_action = QAction(
            "Visit Git repository",
            self.window,
            triggered=partial(self.open_url, REPOSITORY_URL),
        )
        menu.addAction(self.visit_repository_action)

    def about(self):
        About(self.window).exec()

    def open_url(self, url, checked=False):
        qurl = QUrl(url)
        if not QDesktopServices.openUrl(qurl):
            QMessageBox.warning(
                self.window, "Open Url", f"Could not open url '{url}'"
            )

    def start_occurence(self):
        self.toggle_occurrence_menu("start")

    def finish_occurence(self):
        self.toggle_occurrence_menu("finish")

    def toggle_occurrence_menu(self, action):
        self.add_occurrence_action.setText(
            "Finish Occurence" if action == "start" else "Start  Occurence"
        )
        self.abort_occurrence_creation_action.setEnabled(action == "start")
        self.delete_occurrence_action.setEnabled(action == "finish")

    def edit_timeline_properties(self):
        self.window.time_pane.selected_timeline.edit_properties()

    def edit_timeline_events(self):
        self.window.time_pane.selected_timeline.edit_events()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
from yaml import (
    load,
    Loader,
)

from .format import (
    FORMAT,
    format_ok,
)


class Data(pd.DataFrame):
    def __init__(self, filename, file_format):
        super().__init__(pd.read_csv(filename, sep=None, engine="python"))
        self.update_format(file_format)

    def update_format(self, file_format):
        for fmt in range(file_format + 1, FORMAT):
            if fmt == 3:
                self.rename(columns={"label": "event"}, inplace=True)


class MetaData(dict):
    def __init__(self, filename, window=None):
        super().__init__(self._load_file(filename))
        if "format" not in self:
            self = None
        else:
            if not format_ok(self["format"], window):
                self = None
            else:
                self.update_format()

    def _load_file(self, fid):
        data = load(fid, Loader=Loader)
        if not isinstance(data, dict):
            raise TypeError(
                "Metadata file does not contain a dict at top level"
            )
        return data

    def update_format(self):
        for fmt in range(self["format"] + 1, FORMAT + 1):
            if fmt == 4:
                filename = self["video"]
                self["video"] = {
                    "filename": filename,
                    "size": None,
                    "sha1sum": None,
                }
